require "math"
require "util"

local modes = require ".modes"
local debug = require ".debug"


-- TOOD: Need to abstract a `get_splitter (entity)` to fetch the
--   `splitter` from `global.splitters`

-- TODO: change modname to moddir for __hacked-splitters__


-- For configuration changes.
local modname = "hacked-splitters"

-- Belts for the splitters, used to determine how frequently to check
-- for items.
local hacked_splitter_belts = {
   ["hacked-splitter"] = "transport-belt",
   ["hacked-fast-splitter"] = "fast-transport-belt",
   ["hacked-express-splitter"] = "express-transport-belt",

   -- Bob's Logistics mod
   ["hacked-green-splitter"] = "green-transport-belt",
   ["hacked-purple-splitter"] = "purple-transport-belt",
}


script.on_init (function() initialize() end)
script.on_configuration_changed (
   function (data)
      local rebuild = false
      local changes = data.mod_changes

      if changes then
         dprint ('CONFIG:', serpent.line (changes))
      end

      if changes and changes[modname] then
         local mod = changes[modname]
         if mod.old_version == nil then
            -- It's being added. Create skeleton data structures.
            initialize()
         elseif mod.old_version ~= mod.new_version then
            -- Migrations not covered in `migrations/*.lua'
            print (modname, "migration", mod.old_version, mod.new_version)

            -- Until one of major.minor.micro exceeds "9", we can
            -- lexigraphically compare.
            if mod.new_version >= "1.0.3" then
               print (modname, "migrating", "new >= 1.0.3")
               global.belt_cache = nil
               rebuild = true
            end

            if mod.old_version < "1.1.0" then
               print (modname, "migrating", "old < 1.1.0")

               -- There was a problem with Upgrade Planner where
               -- updating items that tracked state on buit/mined
               -- would not trigger those events. Try and fix this.
               --
               -- This also has the benefit of updating all the
               -- hacked splitters to the new data format.
               for i, surface in pairs (game.surfaces) do
                  local s = surface.find_entities_filtered{type="splitter"}
                  for i, entity in pairs (s) do
                     if entity.valid and is_splitter (entity) then
                        -- Use the hammer, recreate them all.
                        local freq = update_frequency (entity.name)
                        local key = poskey (entity)
                        global.splitters[freq][key] = create_splitter (entity)
                     end
                  end
               end
            end
         end
      end

      if changes and changes["base"] then
         rebuild = true
      end

      if changes and changes["boblogistics"] then
         rebuild = true
      end

      if rebuild then
         rebuild_frequency_indices()
         dprint ('CONFIG:', serpent.line (global))
      end
   end
)


script.on_event (
   defines.events.on_player_created,
   function (event)
      if debug.enabled then
         dprint (event.tick, "DEBUG:", "enabled")
         local player = game.players[event.player_index]
         debug.all_the_things (game, player)
      end
   end
)


script.on_event (
   defines.events.on_tick,
   function (event)
      process_splitters (event.tick)
   end
)


script.on_event (
   {
      defines.events.on_built_entity,
      defines.events.on_robot_built_entity
   },
   function (event)
      local entity = event.created_entity
      if is_splitter (entity) then
         -- Now we'll do the same-ol' creation logic for all of our
         -- splitters irrespective (that's not a word!!) of type.
         local freq = update_frequency (entity.name)
         local key = poskey (entity)

         dprint (event.tick, "BUILT:", key, entity.name)
         global.splitters[freq] = global.splitters[freq] or {}
         global.splitters[freq][key] = create_splitter (entity)
      end
   end
)


script.on_event (
   {
      defines.events.on_entity_died,
      defines.events.on_preplayer_mined_item,
      defines.events.on_robot_pre_mined
   },
   function (event)
      local entity = event.entity
      if is_splitter (entity) then
         local splitter, freq, key = find_splitter (entity)

         dprint (event.tick, "MINED:", key, entity.name)
         delete_splitter (freq, key)
      end
   end
)


script.on_event (
   defines.events.on_entity_settings_pasted,
   function (event)
      local src = event.source
      local dst = event.destination

      if is_splitter (src) and is_splitter (dst) then
         src_splitter = find_splitter (src)
         dst_splitter = find_splitter (dst)

         dst_splitter.mode = util.table.deepcopy (src_splitter.mode)
      end
   end
)


-- Handle the keybinds.
function keybind (event, fn)
   local player = game.players[event.player_index]
   local entity = player.selected
   if (entity and is_splitter (entity)
          and player.can_reach_entity (entity))
   then
      local msgs = fn (find_splitter (entity))
      for i, m in pairs (msgs) do
         player.print (m)
      end
   end
end


script.on_event (
   "hacked-splitter-status",
   function(event) keybind (event, modes.keybind_status) end)
script.on_event (
   "hacked-splitter-input-mode",
   function (event) keybind (event, modes.keybind_input) end)
script.on_event (
   "hacked-splitter-output-mode",
   function (event) keybind (event, modes.keybind_output) end)


function initialize()
   -- The hacked splitters that have been placed in the world.
   global.splitters = {}
end


function create_splitter (entity)
   --[[--
      The state data for a splitter. The fields are:
        entity   - the hacked splitter game object
        mode     - the mode type and settings
   --]]--
   local splitter = {
      entity=entity,
      mode=modes.default()
   }
   return splitter
end


function delete_splitter (freq, key)
   if global.splitters[freq] then
      global.splitters[freq][key] = nil
   end
end


function find_splitter (entity)
   local freq = update_frequency (entity.name)
   local key = poskey (entity)
   return global.splitters[freq][key], freq, key
end


function poskey (entity)
   -- Create our own location index into the `splitters` table.  It
   -- requires the surface since multiple surfaces can have the same
   -- position coordinates.
   local name = entity.surface.name
   local pos = entity.position

   return name .. ":" .. tostring (pos.x) .. ":" .. tostring (pos.y)
end


function first (t)
   -- Simple routine to give only the first item
   -- found in a table.
   for k, v in pairs (t) do return k, v end
end


function process_splitter (tick, splitter, key)
   local first = first
   local entity = splitter.entity

   -- We create the iterator here because the number of outbound lanes
   -- is finite. There's no point in looping on all of them for each
   -- inbound lane.
   local output_iter = modes.transport_iter (splitter, modes.SIDE_OUTPUT)

   for input in modes.transport_iter (splitter, modes.SIDE_INPUT) do
      -- The hack to the splitter created a bit of the Schrodinger
      -- Splitter--you really can't tell which item on the
      -- transport line will come out first.
      local tin = entity.get_transport_line (input)
      local item, count = first (tin.get_contents())

      if item then
         for output in output_iter do
            -- Here's the actual act. Everything else was foreplay.
            local tout = entity.get_transport_line (output)

            -- Inserting half-way seems like a good performance
            -- trade-off. It works in most situtations. When changing
            -- the insert logic to "for i=0,.5,.001" during testing,
            -- often it would be able to insert in the 0.3 to 0.4
            -- range unless it was purely a cross-side balance (e.g.,
            -- 6-8 blocked, balancing in/out of 4/5), in which case it
            -- could insert it at 0 since the normal splitter would
            -- never send to that line.

            dprint (tick, 'in/out', tostring(input)..'/'..tostring(output))

            if tout.insert_at (0.5, {name=item}) then
               tin.remove_item ({name=item})

               -- Update the last lines that were successfully used.
               modes.transport_update (splitter, input, output)
               dprint (tick, 'in/out', tostring(input)..'/'..tostring(output), '*')

               break
            end
         end

         -- If we fail to insert, it's not the end of the world.  To
         -- keep a compressed belt, the default splitter logic will
         -- take over. This happens when an outbound belt is
         -- completely stopped.
      end
   end
end


function process_splitters (tick)
   for freq, splitters in pairs (global.splitters) do
      if tick % freq == 0 then
         for key, splitter in pairs (splitters) do
            -- I1: Somehow the entity wasn't valid. I wanted to avoid
            -- these checks in the main-loop, but someone reported a
            -- case where `entity` was invalid. Now that I've played
            -- with upgrade-planner, I think this was the cause.
            if not splitter.entity.valid then
               dprint (tick, "splitter", key, "invalid!")
               delete_splitter (freq, key)
            else
               process_splitter (tick, splitter, key)
            end
         end
      end
   end
end


function is_splitter (entity)
   -- Or more specifically, a hacked splitter.
   if entity.type ~= "splitter" then
      return false
   else
      return hacked_splitter_belts[entity.name]
   end
end


function calc_belt_ticks (belt)
   --[[--
      To my knowledge, you can't interrogate the recipes at this point
      nor can you access the `.speed` field of the splitter prototype
      since it isn't exposed. Instead, we *assume* the proper speed is
      paired with a belt type.

      The time an item is spent in the splitter unit appears to be:
        calc_time = 1 / speed
          ; basic=32, fast=16, express=10.667
        actual_time = math.floor (calc_time)
          ; basic=32, fast=16, express=10

      This means an express splitter is faster than an express belt.

      We have to make a decision to balance the item before it
      transitions to the output line automatically--i.e., before the
      default splitter logic makes the decision for us. So, how do we
      determine this programmatically?

      Given a fixed point on the belt, we know how many ticks it takes
      for an item to pass by that point. We use the unit distance to
      calculate this. As of Factorio 0.13.3, this is 0.28125 units.

        tick_interval = 0.28125 / speed
          ; basic=9, fast=4.5, express=3

      However, in-game with the fast splitter, the ticks between when
      an item enters the inbound-side and the next item is 4. Likewise
      for when it enters the outbound-side. Cautiously, I'll say the
      tick interval for the first item on inbound and outbound is:

         tick_interval = math.floor (0.28125 / speed)

   --]]--

   local inbound = nil
   if belt ~= nil then
      local entity = game.entity_prototypes[belt]
      if entity and entity.type == "transport-belt" then
         local tick_interval = 0.28125 / entity.belt_speed
         -- Need to check at half the time it takes to not miss items.
         inbound = math.max (math.floor (tick_interval / 2), 1)
      end
   end

   return inbound
end


function update_frequency (name)
   -- How frequently `on_tick` is run for this splitter.
   return calc_belt_ticks (hacked_splitter_belts[name])
end


function rebuild_frequency_indices()
   -- Whenever a mod changes that has splitters, they may have updated
   -- the belt speed. The `freq` index might no longer run fast (or
   -- slow) enough for the change. This rebuilds the frequency
   -- table to reflect those changes.
   new = {}

   for freq, splitters in pairs (global.splitters) do
      for key, splitter in pairs (splitters) do
         local entity = splitter.entity
         if entity.valid then
            local update = update_frequency (entity.name)
            if update ~= nil then
               new[update] = new[update] or {}
               new[update][key] = splitter
            end
         end
      end
   end

   global.splitters = new
end
