require 'math'


local modes = {}

modes.MODE_DEFAULT = 1  -- The Original++

modes.SIDE_INPUT = 1    -- Input side of the splitter
modes.SIDE_OUTPUT = 2   -- Output side of the splitter

modes.SIDE_DESC = {
   [modes.SIDE_INPUT] = "input",
   [modes.SIDE_OUTPUT] = "output"
}

modes.PRIORITY_DESC = {
   [0] = "right",
   [50] = "balanced",
   [100] = "left"
}


-- Gotta have a little fun.
function console (splitter, command, stdout)
   local entity = splitter.entity
   local x, y = entity.position.x, entity.position.y
   local hostname = string.format ("hksplt%+.1f%+.1f", x, y)

   local msg = { string.format ("root@%s$ %s", hostname, command) }
   if stdout then table.insert (msg, stdout) end
   return msg
end


function modes.default()
   local transport_line = defines.transport_line

   return {
      options = modes.MODE_DEFAULT,
      settings = {
         [modes.MODE_DEFAULT] = {
            [modes.SIDE_INPUT] = {
               line = transport_line.left_line,
               priority = 50
            },
            [modes.SIDE_OUTPUT] = {
               line = transport_line.left_split_line,
               priority = 50
            }
         }
      }
   }
end


function modes.keybind_status (splitter)
   if splitter == nil then return end

   local settings = splitter.mode.settings[modes.MODE_DEFAULT]
   local setin = settings[modes.SIDE_INPUT]
   local setout = settings[modes.SIDE_OUTPUT]

   local f = function (priority)
      local str = ""

      if priority == 0 then
         str = "right"
      elseif priority == 50 then
         str = "balanced"
      else
         str = "left"
      end

      return str
   end

   local inp = f (setin.priority)
   local out = f (setout.priority)

   return console (splitter, "cat /proc/kern_pri/status", inp.." / "..out)
end


function keybind_inout (splitter, side)
   if splitter == nil then return end
   local mode = splitter.mode

   if mode.options == modes.MODE_DEFAULT then
      local setting = mode.settings[mode.options][side]

      -- Cycle it through 50-100-0 from left's perspective.
      setting.priority = (setting.priority + 50) % 150

      return console (
         splitter,
         string.format ("echo %s > /proc/kern_pri/%s",
                        modes.PRIORITY_DESC[setting.priority]
                        , modes.SIDE_DESC[side])
      )
   end
end


function modes.keybind_input (splitter)
   return keybind_inout (splitter, modes.SIDE_INPUT)
end


function modes.keybind_output (splitter)
   return keybind_inout (splitter, modes.SIDE_OUTPUT)
end


-- A little state machine indicating the order of transport
-- lines--both input and output--to check based on the last line
-- checked.
local transport_line = defines.transport_line
local transport_fsm = {
   -- Input lines.
   [ transport_line.left_line ] = transport_line.secondary_left_line,
   [ transport_line.secondary_left_line ] = transport_line.right_line,
   [ transport_line.right_line ] = transport_line.secondary_right_line,
   [ transport_line.secondary_right_line ] = transport_line.left_line,

   -- Output lines.
   [ transport_line.left_split_line ] = transport_line.secondary_left_split_line,
   [ transport_line.secondary_left_split_line ] = transport_line.right_split_line,
   [ transport_line.right_split_line ] = transport_line.secondary_right_split_line,
   [ transport_line.secondary_right_split_line ] = transport_line.left_split_line
}


function transport_default_iter (initial)
   -- A special property of this iterator is that it will iterate
   -- through all lines on a side of the splitter, once.
   local transport_fsm = transport_fsm
   local curr = nil
   return function()
      if curr ~= initial then
         curr = curr or initial
         curr = transport_fsm[curr]
         return curr
      end
   end
end


-- This would be easier if I math'ed it. I know that the lines are
-- I:12 34 O:56 78, but I'll use the defines in case they ever change
-- the values (yeah... right).
local transport_priority_fsms = {
   [modes.SIDE_INPUT] = {
      [0] = {
         transport_line.secondary_left_line,
         transport_line.secondary_right_line,
         transport_line.left_line,
         transport_line.right_line
      },
      [100] = {
         transport_line.left_line,
         transport_line.right_line,
         transport_line.secondary_left_line,
         transport_line.secondary_right_line
      },
   },
   [modes.SIDE_OUTPUT] = {
      [0] = {
         transport_line.secondary_right_split_line,
         transport_line.secondary_left_split_line,
         transport_line.right_split_line,
         transport_line.left_split_line,
      },
      [100] = {
         transport_line.right_split_line,
         transport_line.left_split_line,
         transport_line.secondary_right_split_line,
         transport_line.secondary_left_split_line
      }
   }
}


function transport_priority_iter (side, priority)
   -- Hacked splitters use a priority random distribution mode whenever
   -- they are not equally balanced. This makes state management a little
   -- easier as we are not concerned with the last priority line *and* the
   -- last secondary line.
   local a = math.random (1, 2)
   local b = math.random (1, 2)

   local order = transport_priority_fsms[side][priority]
   local dist = { order[a], order[a % 2 + 1], order[b + 2], order[b % 2 + 3] }

   local cnt = 0
   return function()
      cnt = cnt + 1
      return dist[cnt]
   end
end


function modes.transport_iter (splitter, side)
   local mode = splitter.mode
   local settings = mode.settings[mode.options][side]

   if mode.options == modes.MODE_DEFAULT then
      local priority = settings.priority

      if priority == 50 then
         return transport_default_iter (settings.line)
      else
         return transport_priority_iter (side, priority)
      end
   end
end


function modes.transport_update (splitter, input, output)
   local mode = splitter.mode
   local settings = mode.settings[mode.options]

   if mode.options == modes.MODE_DEFAULT then
      settings[modes.SIDE_INPUT].line = input
      settings[modes.SIDE_OUTPUT].line = output
   end
end


return modes
